#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define IPA(x) ((htonl(x)&0xff000000)>>24)
#define IPB(x) ((htonl(x)&0xff0000)>>16)
#define IPC(x) ((htonl(x)&0xff00)>>8)
#define IPD(x) (htonl(x)&0xff)

/*
 * default number of latency timings to take
 */
#define DEFNUM 9

struct results {
	struct timeval res_tim;
	struct results *res_nxt;
} *restop;

int timedout, inted;

void
intsig()
{
	inted=1;
}

void
alrm()
{
	timedout=1;
}

main(argc, argv)
char **argv;
{
	char *rbuf, *wbuf, *p, *sendstr=NULL, *malloc(), *strchr();
	int size=1, number=DEFNUM, vflg=0, qflg=0, timeout=0, portno=7;
	int n, fail, ok, getmid=1;
	int fd, wlen, rlen, rsplcnt, wsplcnt;
	int optch;
	extern int optind;
	extern char *optarg;
	struct timeval tstart, tend, ttotal, tmin, tmax;
	struct results *resptr, *resnew, *resprv;
	struct hostent *hp;
	struct sockaddr_in sin;

	while ((optch=getopt(argc, argv, "c:n:s:t:vq")) != -1)
		switch (optch) {
		case 'c':
			sendstr=optarg;
			break;
		case 'n':
			number=atoi(optarg);
			break;
		case 's':
			size=atoi(optarg);
			break;
		case 't':
			timeout=atoi(optarg);
			break;
		case 'q':
			qflg++;
			break;
		case 'v':
			vflg++;
			break;
		default:
			goto usage;
		}
	if (optind != argc-1) {
usage:
		fprintf(stderr, "Usage: %s [-v[v]] [-q] [-s<size>] [-n<number>] [-c<pattern>] [-t<timeout>] <host>[:<port>]\n", argv[0]);
		fprintf(stderr, "\t-v\t\tverbose (-vv very verbose)\n");
		fprintf(stderr, "\t-q\t\tquiet - do not show each result\n");
		fprintf(stderr, "\t-s <size>\tnumber of bytes to send (default 1)\n");
		fprintf(stderr, "\t-n <number>\tnumber of latency tests (default %d)\n", DEFNUM);
		fprintf(stderr, "\t-c <pattern>\trepeated string to send (default random)\n");
		fprintf(stderr, "\t-t <timeout>\ttimeout for response (default no timeout)\n");
		exit(1);
	}

	/*
	 * Create read and write buffers
	 */
	if ((rbuf=malloc(size)) == NULL || (wbuf=malloc(size+1)) == NULL) {
		fprintf(stderr, "%s: Cannot create buffer of size %d\n", argv[0], size);
		exit(1);
	}
	/*
	 * fill write buffer with string to send
	 */
	if (sendstr) {
		for (p=sendstr, wlen=0; wlen < size; wlen++) {
			wbuf[wlen] = *p++;
			if (*p == '\0')
				p=sendstr;
		}
	} else {
		srand(getpid());
		for (wlen=0; wlen < size; wlen++)
			wbuf[wlen] = rand()%95+' ';
	}
	wbuf[size] = '\0';
	if (vflg > 1)
		printf("Buffer: %s\n", wbuf);

	/*
	 * check for alternate port number
	 */
	if (p=strchr(argv[optind], ':')) {
		portno=atoi(&p[1]);
		*p='\0';
	}

	/*
	 * Lookup hostname (or IP) given
	 */
	if (vflg)
		printf("Looking up hostname: %s\n", argv[optind]);
	bzero((char *) &sin, sizeof(sin));
	if ((sin.sin_addr.s_addr=inet_addr(argv[optind])) == -1 || sin.sin_addr.s_addr == 0) {
		if ((hp=gethostbyname(argv[optind])) == NULL) {
			fprintf(stderr, "%s: Can't find host \"%s\"\n", argv[0], argv[1]);
			exit(1);
		}
		bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
	}
	if (vflg)
		printf("Remote host: %d.%d.%d.%d\n", IPA(sin.sin_addr.s_addr), IPB(sin.sin_addr.s_addr), IPC(sin.sin_addr.s_addr), IPD(sin.sin_addr.s_addr));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(portno);

	/*
	 * Initialize ok and fail count, and total time
	 */
	n=0;
	ok=0;
	fail=0;
	ttotal.tv_sec=0;
	ttotal.tv_usec=0;

reopen:
	/*
	 * Make connection
	 */
	if (vflg)
		printf("Connecting to remote echo port: TCP#%d\n", portno);
	if ((fd=socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		exit(1);
	}
	if (connect(fd, &sin, sizeof(sin)) == -1) {
		perror("connect");
		exit(1);
	}

	/*
	 * setup user interrupt to exit, but report findings
	 */
	signal(SIGINT, intsig);

	if (n == 0 && vflg)
		printf("Starting %d write/read loop%s\n", number, number == 1?"":"s");

	for (; !inted && n < number; n++) {
		if (timeout) {
			signal(SIGALRM, alrm);
			timedout=0;
			alarm(timeout);
		}
		/*
		 * Start timer
		 */
		gettimeofday(&tstart, NULL);
		rlen=size;
		p=wbuf;
		wsplcnt=0;
		/*
		 * write buffer until empty
		 */
		while (rlen) {
			if ((wlen=write(fd, p, rlen)) == -1) {
				if (inted) {
					printf("User interrupt\n");
					break;
				}
				if (qflg == 0) {
					if (timedout)
						printf("Write timed out\n");
					else
						perror("write");
				}
				break;
			}
			p += wlen;
			rlen -= wlen;
			wsplcnt++;
		}
		if (rlen) {
			if (timeout)
				alarm(0);
			if (vflg)
				printf("Closing connection\n");
			close(fd);
			fail++;
			n++;
			goto reopen;
		}
		wlen=size;
		p=rbuf;
		rsplcnt=0;
		/*
		 * read into buffer until full
		 */
		while (wlen) {
			if ((rlen=read(fd, p, wlen)) == -1) {
				if (inted) {
					printf("User interrupt\n");
					break;
				}
				if (qflg == 0) {
					if (timedout)
						printf("Read timed out\n");
					else
						perror("read");
				}
				break;
			}
			p += rlen;
			wlen -= rlen;
			rsplcnt++;
		}
		if (timeout)
			alarm(0);
		if (wlen) {
			if (vflg)
				printf("Closing connection\n");
			close(fd);
			fail++;
			n++;
			goto reopen;
		}
		/*
		 * Stop timer
		 */
		gettimeofday(&tend, NULL);

		/*
		 * compare write with read
		 */
		if (memcmp(rbuf, wbuf, size) != 0) {
			if (qflg == 0)
				printf("written != read\n");
			if (vflg)
				printf("Closing connection\n");
			close(fd);
			fail++;
			n++;
			goto reopen;
		}
		ok++;

		/*
		 * calculate difference
		 */
		tend.tv_sec -= tstart.tv_sec;
		if (tend.tv_usec < tstart.tv_usec) {
			tend.tv_usec += 1000000;
			tend.tv_sec  -= 1;
			tend.tv_usec -= tstart.tv_usec;
		} else
			tend.tv_usec -= tstart.tv_usec;

		/*
		 * report findings
		 */
		if (qflg == 0)
			printf("%d byte%s written in %d chunk%s, read in %d chunk%s, delay %d.%06d\n", size, size == 1?"":"s", wsplcnt, wsplcnt == 1?"":"s", rsplcnt, rsplcnt == 1?"":"s", tend.tv_sec, tend.tv_usec);

		/*
		 * find max and min
		 */
		if (ttotal.tv_sec == 0 && ttotal.tv_usec == 0) {
			/*
			 * first time
			 */
			tmin.tv_sec = tend.tv_sec;
			tmin.tv_usec = tend.tv_usec;
			tmax.tv_sec = tend.tv_sec;
			tmax.tv_usec = tend.tv_usec;
		} else {
			/*
			 * bigger than max so far?
			 */
			if (tend.tv_sec > tmax.tv_sec || (tend.tv_sec == tmax.tv_sec && tend.tv_usec > tmax.tv_usec)) {
				tmax.tv_sec = tend.tv_sec;
				tmax.tv_usec = tend.tv_usec;
			}
			/*
			 * smaller than min so far?
			 */
			if (tend.tv_sec < tmin.tv_sec || (tend.tv_sec == tmin.tv_sec && tend.tv_usec < tmin.tv_usec)) {
				tmin.tv_sec = tend.tv_sec;
				tmin.tv_usec = tend.tv_usec;
			}
		}

		/*
		 * add this time to total
		 */
		ttotal.tv_sec += tend.tv_sec;
		ttotal.tv_usec += tend.tv_usec;
		/*
		 * carry microseconds in seconds
		 */
		if (ttotal.tv_usec > 1000000) {
			ttotal.tv_sec += ttotal.tv_usec/1000000;
			ttotal.tv_usec %= 1000000;
		}

		if (getmid) {
			/*
			 * add to list of results
			 */
			if ((resnew=(struct results *)malloc(sizeof(struct results))) == NULL) {
				getmid=0;
				continue;
			}
			resnew->res_tim.tv_sec=tend.tv_sec;
			resnew->res_tim.tv_usec=tend.tv_usec;
			/*
			 * find place to insert
			 */
			resprv=NULL;
			for (resptr=restop; resptr; resptr=resptr->res_nxt) {
				if (tend.tv_sec < resptr->res_tim.tv_sec || (tend.tv_sec == resptr->res_tim.tv_sec && tend.tv_usec < resptr->res_tim.tv_usec))
					break;
				resprv=resptr;
			}
			if (resprv == NULL) {
				/*
				 * add at start
				 */
				resnew->res_nxt=restop;
				restop=resnew;
			} else {
				/*
				 * insert here
				 */
				resprv->res_nxt=resnew;
				resnew->res_nxt=resptr;
			}
		}
	}

	/*
	 * report findings on entire session
	 */
	if (ok == 0) {
		printf("Total packet loss\n");
		exit(0);
	}
	printf("%d/%d ok (%d%%)\n", ok, ok+fail, ok*100/(ok+fail));
	size = ttotal.tv_sec % ok;
	ttotal.tv_usec += size*1000000;
	ttotal.tv_sec /= ok;
	ttotal.tv_usec /= ok;

	printf("min=%d.%06d ", tmin.tv_sec, tmin.tv_usec);
	printf("ave=%d.%06d ", ttotal.tv_sec, ttotal.tv_usec);
	printf("max=%d.%06d ", tmax.tv_sec, tmax.tv_usec);

	if (getmid) {
		resptr=restop;
		for (n=(ok-1)/2; n--;)
			resptr=resptr->res_nxt;
		if (ok%2 == 0) {
			printf("mid=%d.%06d/", resptr->res_tim.tv_sec, resptr->res_tim.tv_usec);
			resptr=resptr->res_nxt;
		}
		printf("mid=%d.%06d\n", resptr->res_tim.tv_sec, resptr->res_tim.tv_usec);
	} else
		printf("mid=Memory-error\n");

	/*
	 * maybe close() should be called - let exit do it
	 */
	exit(0);
}
